"use strict";
var __spreadArrays = (this && this.__spreadArrays) || function () {
    for (var s = 0, i = 0, il = arguments.length; i < il; i++) s += arguments[i].length;
    for (var r = Array(s), k = 0, i = 0; i < il; i++)
        for (var a = arguments[i], j = 0, jl = a.length; j < jl; j++, k++)
            r[k] = a[j];
    return r;
};
Object.defineProperty(exports, "__esModule", { value: true });
var lodash_1 = require("lodash");
var validation_error_1 = require("../exceptions/validation.error");
var GraphqlGenerator = /** @class */ (function () {
    function GraphqlGenerator(requestTypeName) {
        this.requestTypeName = requestTypeName;
    }
    GraphqlGenerator.validateRequests = function (requests) {
        requests.forEach(GraphqlGenerator.validateRequest);
    };
    GraphqlGenerator.validateRequest = function (request) {
        var fragmentParams = request.fragmentParams, fragmentValues = request.fragmentValues, fragmentName = request.fragmentName;
        if (Array.isArray(fragmentParams) && typeof (fragmentValues) === 'object') {
            var valueNames_1 = Object.keys(fragmentValues);
            if (!valueNames_1.length) {
                throw new validation_error_1.ValidationError("request with name: " + fragmentName + " has parameters but no values");
            }
            var requestParamNames = fragmentParams.map(function (param) { return param.name; });
            var missingValueNames = requestParamNames.filter(function (paramName) {
                if (valueNames_1.indexOf(paramName) === -1)
                    return true;
                if (!fragmentValues[paramName])
                    return true;
            });
            if (missingValueNames.length) {
                throw new Error("The following values are missing from the request:\n" + missingValueNames.join('\n-'));
            }
        }
    };
    GraphqlGenerator.collectRequestsParameters = function (requests) {
        return requests.reduce(function (params, request) {
            return params.concat(request.fragmentParams || []);
        }, []);
    };
    GraphqlGenerator.generateParameterAlias = function (param) {
        var name = param.name, alias = param.alias;
        if (!name)
            throw new validation_error_1.ValidationError('Missing required param name');
        return alias ? alias : "$" + name;
    };
    GraphqlGenerator.prototype.generateRequestHeader = function (requests) {
        var params = GraphqlGenerator.collectRequestsParameters(requests);
        var requestHeader = this.requestTypeName;
        if (params.length)
            requestHeader += "(" + GraphqlGenerator.collectHeaderParams(params) + ")";
        return requestHeader;
    };
    GraphqlGenerator.collectHeaderParams = function (params) {
        return params.map(GraphqlGenerator.collectHeaderParam).join(',');
    };
    GraphqlGenerator.collectHeaderParam = function (param) {
        if (!param.type)
            throw new validation_error_1.ValidationError('Missing required param type');
        var useAlias = GraphqlGenerator.generateParameterAlias(param);
        return [useAlias, param.type].join(':');
    };
    GraphqlGenerator.prototype.generateRequestFragments = function (requests) {
        return requests.map(this.generateRequestFragment.bind(this)).join('\n');
    };
    GraphqlGenerator.prototype.generateRequestFragment = function (request) {
        var fragmentString = '';
        var fragmentName = request.fragmentName, fragmentParams = request.fragmentParams, fragmentFields = request.fragmentFields;
        var fragmentHeader = GraphqlGenerator.generateFragmentHeader(fragmentName, fragmentParams);
        fragmentString += fragmentHeader;
        if (fragmentFields) {
            if (!Array.isArray(fragmentFields)) {
                throw new Error('Fields must be an array');
            }
            if (fragmentFields.length) {
                var fragmentBody = this.generateFragmentBody(fragmentFields);
                fragmentString += "{" + fragmentBody + "}";
            }
        }
        return fragmentString;
    };
    GraphqlGenerator.generateFragmentHeader = function (name, params) {
        if (params === void 0) { params = []; }
        var fragmentHeader = name;
        if (params.length)
            fragmentHeader += "(" + GraphqlGenerator.generateFragmentHeaderParams(params) + ")";
        return fragmentHeader;
    };
    GraphqlGenerator.generateFragmentHeaderParams = function (params) {
        return params.map(GraphqlGenerator.generateFragmentHeaderParam).join(',');
    };
    GraphqlGenerator.generateFragmentHeaderParam = function (param) {
        var useAlias = GraphqlGenerator.generateParameterAlias(param);
        return [param.name, useAlias].join(':');
    };
    GraphqlGenerator.prototype.generateFragmentBody = function (fields) {
        var fieldsObject = this.collectFieldsAsObject(fields);
        var fragmentBodyComponents = this.generateFragmentFieldsString(fieldsObject);
        return fragmentBodyComponents.join(' ');
    };
    GraphqlGenerator.prototype.collectFieldsAsObject = function (fields) {
        return fields.reduce(function (object, field) {
            lodash_1.set(object, field, field);
            return object;
        }, {});
    };
    GraphqlGenerator.prototype.generateFragmentFieldsString = function (fieldsObject) {
        var _this = this;
        return Object.keys(fieldsObject).reduce(function (params, key) {
            var value = fieldsObject[key];
            if (typeof (value) === 'object') {
                params.push.apply(params, __spreadArrays([key, '{'], _this.generateFragmentFieldsString(value), ['}']));
            }
            else {
                params.push(key);
            }
            return params;
        }, []);
    };
    GraphqlGenerator.prototype.generateRequestString = function (requests) {
        GraphqlGenerator.validateRequests(requests);
        var queryHeader = this.generateRequestHeader(requests);
        var queryFragments = this.generateRequestFragments(requests);
        return queryHeader + "{\n" + queryFragments + "}";
    };
    return GraphqlGenerator;
}());
exports.default = GraphqlGenerator;
